/*
 * timemgt.c
 *
 *  Created on: Apr 28, 2021
 *      Author: Florian
 */

#include "timemgt.h"
#include "string.h"

/* Local static variables */
tTimeMgt_Channel Channels_tA[TIM_ChannelNumberMax_U8];
static tTimeMgt_List list_T;
//static tTimeMgt_Channel *list_T;

static uint8_t channels_cnt_u8;
static uint32_t id_cnt_u32; // Free running counter - never decremented

/* Local static functions */
BOOL CompareStrings (char *string1, char *string2)
{
	BOOL ret_b = FALSE;

	return ret_b;
}

/* Functions */
void TimeMgt_Init (void)
{
	/*tTimeMgt_Channel *channel = malloc(sizeof(tTimeMgt_Channel));

	channel->next_p = NULL;
	channel->enable_b = FALSE;
	list_T->head = channel;*/
	list_T.head = NULL;

	channels_cnt_u8 = 0U;
	id_cnt_u32 = 0U;
}

void TimeMgt_Start (void)
{

}

void TimeMgt_Stop(void)
{
	uint8_t cnt = 0U;

	for (cnt=0U; cnt < TIM_ChannelNumberMax_U8; cnt++)
	{
		TimeMgt_DisableCyclicEvent(cnt);
		TimeMgt_RemoveCyclicEvent(cnt);
	}
}

BOOL TimeMgt_AddCyclicEvent(uint8_t channel_u8, pTimeMgtCallback callback_p, uint32_t timeout_u32, char *tag_u8p, uint8_t *id_u32p)
{
	BOOL ret_b = FALSE;

	// Tab managed
	if ((channel_u8 < TIM_ChannelNumberMax_U8) && (callback_p != NULL))
	{
		Channels_tA[channel_u8].callback_p = callback_p;
		Channels_tA[channel_u8].timeout_u32 = timeout_u32;
		Channels_tA[channel_u8].counter_u32 = 0U;
	}

	// Chained list managed
	if (channels_cnt_u8 < cTimeMgtChannelMax)
	{
		tTimeMgt_Channel *channel = malloc(sizeof(tTimeMgt_Channel));

		if (list_T.head == NULL) // If no object present in the list
		{
			channel->next_p = NULL;
			list_T.head = channel;
		}
		else // If objects present in the list
		{
			channel->next_p = list_T.head;
			list_T.head = channel;
		}
		channel->enable_b = TRUE;
		channel->counter_u32 = 0U;
		channel->callback_p = callback_p;
		channel->timeout_u32 = timeout_u32;

		// Increment channel counter
		channels_cnt_u8++;

		// Assign unique ID
		*id_u32p = id_cnt_u32;
		channel->id_u32 = id_cnt_u32;
		id_cnt_u32++;

		// Assign tag
		uint8_t cnt_u8 = 0;
		BOOL end_of_line = FALSE;

		for (cnt_u8 = 0; cnt_u8 < cTimeMgtTagStringMaxLength; cnt_u8++)
		{
			if (*(tag_u8p + cnt_u8) == (char)0x00)
			{
				end_of_line = TRUE;
			}

			if (end_of_line == FALSE)
			{
				channel->tag_u8a[cnt_u8] = *(tag_u8p + cnt_u8);
			}
			else
			{
				channel->tag_u8a[cnt_u8] = (char)0x00;
			}

		}

		ret_b = TRUE;
	}

	return (ret_b);
}

BOOL TimeMgt_SearchEventById(uint32_t id_u32, tTimeMgt_Channel** prev, tTimeMgt_Channel** match)
{
	BOOL ret_b = FALSE;

	if(list_T.head != NULL) // If list is not empty
	{
		tTimeMgt_Channel *channel = list_T.head;

		if (channel->id_u32 == id_u32) // Specific test for head
		{
			*match = channel;
			*prev = NULL;
			ret_b = TRUE;
		}
		else  // If more than one element in list
		{
			do
			{
				if (channel->next_p != NULL)
				{
					if (channel->next_p->id_u32 == id_u32)
					{
						*match = (channel->next_p);
						*prev = channel;
						ret_b = TRUE;
						break;
					}
				}

				channel = channel->next_p;
			}
			while (channel != NULL);
		}
	}
	else
	{
		*prev = NULL;
		*match = NULL;
	}

	return (ret_b);
}

void TimeMgt_RemoveCyclicEvent(uint8_t channel_u8)
{
	// Tab managed
	if ((channel_u8 < TIM_ChannelNumberMax_U8) && (Channels_tA[channel_u8].callback_p != NULL))
	{
		Channels_tA[channel_u8].callback_p = NULL;
		Channels_tA[channel_u8].timeout_u32 = 0U;
		Channels_tA[channel_u8].counter_u32 = 0U;
		Channels_tA[channel_u8].enable_b = FALSE;
	}

	// Chained list managed
	if(list_T.head != NULL)
	{
		tTimeMgt_Channel *channel = list_T.head;
		list_T.head = list_T.head->next_p;
		free(channel);
	}
}

BOOL TimeMgt_RemoveCyclicEventById(uint32_t id_u32)
{
	BOOL ret_b = FALSE;

	tTimeMgt_Channel *match;
	tTimeMgt_Channel *prev;

	if (TimeMgt_SearchEventById(id_u32, &prev, &match) == TRUE) // Search for the object
	{
		if (prev == NULL) // Match is HEAD
		{
			tTimeMgt_Channel *channel = list_T.head;
			list_T.head = list_T.head->next_p;
			free(channel);
			ret_b = TRUE;
		}
		else if (match != NULL)
		{
			prev->next_p = match->next_p; // Bypass object to be removed
			free(match); // Free object that must be removed
			ret_b = TRUE;
		}
	}

	return (ret_b);
}

BOOL TimeMgt_RemoveCyclicEventByTag(char *tag_u8p)
{
	BOOL ret_b;

	// TODO : factorise with first / next (Lionel method) with static pointer / rover
	if(list_T.head != NULL) // If list is not empty
	{
		tTimeMgt_Channel *channel = list_T.head;

		if (strncmp(channel->tag_u8a, tag_u8p, 8) != 0U) // Specific test for head
		{
			list_T.head = list_T.head->next_p;
			free(channel);
			ret_b = TRUE;
		}
		else  // If more than one element in list
		{
			do
			{
				if (channel->next_p != NULL)
				{
					if (strncmp(channel->next_p->tag_u8a, tag_u8p, 8) != 0U) // next object is the object to be removed
					{
						tTimeMgt_Channel *channel_next = channel->next_p; // Local copy of the pointer to the next object

						channel->next_p = channel_next->next_p; // Change pointer

						free(channel_next); // Free object that must be removed
						ret_b = TRUE;
					}
				}

				channel = channel->next_p;
			}
			while (channel != NULL);
		}
	}

	return (ret_b);
}

void TimeMgt_EnableCyclicEvent(uint8_t channel_u8)
{
	if (channel_u8 < TIM_ChannelNumberMax_U8)
	{
		if (Channels_tA[channel_u8].callback_p != NULL)
		{
			Channels_tA[channel_u8].enable_b = TRUE;
		}
	}
}

BOOL TimeMgt_EnableCyclicEventById(uint32_t id_u32)
{
	BOOL ret_b = FALSE;

	tTimeMgt_Channel *match;
	tTimeMgt_Channel *prev;

	if (TimeMgt_SearchEventById(id_u32, &prev, &match) == TRUE) // Search for the object
	{
		match->enable_b = TRUE;
	}

	return (ret_b);
}

BOOL TimeMgt_EnableCyclicEventByTag(char *tag_u8p)
{
	BOOL ret_b = FALSE;

	if(list_T.head != NULL) // If list is not empty
	{
		tTimeMgt_Channel *channel = list_T.head;

		if (strncmp(channel->tag_u8a, tag_u8p, 8) != 0U) // Specific test for head
		{
			list_T.head->enable_b = TRUE;
			ret_b = TRUE;
		}
		else  // If more than one element in list
		{
			do
			{
				if (channel->next_p != NULL)
				{
					if (strncmp(channel->next_p->tag_u8a, tag_u8p, 8) != 0U) // next object is the object to be removed
					{
						channel->next_p->enable_b = TRUE;
						ret_b = TRUE;
					}
				}

				channel = channel->next_p;
			}
			while (channel != NULL);
		}
	}

	return (ret_b);
}

void TimeMgt_DisableCyclicEvent(uint8_t channel_u8)
{
	if (channel_u8 < TIM_ChannelNumberMax_U8)
	{
		Channels_tA[channel_u8].enable_b = FALSE;
	}
}

BOOL TimeMgt_DisableCyclicEventById(uint32_t id_u32)
{
	BOOL ret_b = FALSE;

	tTimeMgt_Channel *match;
	tTimeMgt_Channel *prev;

	if (TimeMgt_SearchEventById(id_u32, &prev, &match) == TRUE) // Search for the object
	{
		match->enable_b = FALSE;
	}

	return (ret_b);
}

BOOL TimeMgt_DisableCyclicEventByTag(char *tag_u8p)
{
	BOOL ret_b = FALSE;

	if(list_T.head != NULL) // If list is not empty
	{
		tTimeMgt_Channel *channel = list_T.head;

		if (strncmp(channel->tag_u8a, tag_u8p, 8) != 0U) // Specific test for head
		{
			list_T.head->enable_b = FALSE;
			ret_b = TRUE;
		}
		else  // If more than one element in list
		{
			do
			{
				if (channel->next_p != NULL)
				{
					if (strncmp(channel->next_p->tag_u8a, tag_u8p, 8) != 0U) // next object is the object to be removed
					{
						channel->next_p->enable_b = FALSE;
						ret_b = TRUE;
					}
				}

				channel = channel->next_p;
			}
			while (channel != NULL);
		}
	}

	return (ret_b);
}

void TimeMgt_PeriodicFunction(void)
{
	uint8_t cnt;

	for (cnt=0U; cnt < TIM_ChannelNumberMax_U8; cnt++)
	{
		if (Channels_tA[cnt].enable_b == TRUE)
		{
			if (Channels_tA[cnt].counter_u32 < Channels_tA[cnt].timeout_u32)
			{
				Channels_tA[cnt].counter_u32++;
			}
			else
			{
				Channels_tA[cnt].counter_u32 = 0U;
				// Callback
				if (Channels_tA[cnt].callback_p != NULL)
				{
					Channels_tA[cnt].callback_p();
				}
			}

		}
	}

	tTimeMgt_Channel *channel = list_T.head;

	if (list_T.head != NULL)
	{
		tTimeMgt_Channel *channel = list_T.head;

		while(channel != NULL)
		{
			if (channel->enable_b == TRUE)
			{
				if (channel->counter_u32 < channel->timeout_u32)
				{
					channel->counter_u32++;
				}
				else
				{
					channel->counter_u32 = 0U;
					if (channel->callback_p != NULL)
					{
						channel->callback_p();
					}
				}
			}

			channel = channel->next_p;
		}
	}
}


